"""
Simple Helper Module to provide widely used colors. 

These Colors Include:
- `19` Material Colors (with `14` different shades in each)
- `140` Web/HTML Colors
- `200` Individual Color Palettes (with `5` colors in each)
"""

from . import _corpus_colors
from random import choice
from typing import List, Dict


def get_web_colors():
    """
    ### Get Web Colors
    Yields Web/HTML colors in `dict`.
    ```
    #### dict Structure:
    {
        'color':"AliceBlue",
        'hex':"#F0F8FF"
    }
    ```
    """
    for color in _corpus_colors.web_colors:
        yield color


def get_palettes():
    """
    ### Get Palettes
    Yields a palette per `list`.
    ```
    #### List Structure:
    [
        "#69D2E7",
        "#A7DBD8",
        "#E0E4CC",
        "#F38630",
        "#FA6900"
    ]
    ```
    """
    for palette in _corpus_colors.palettes:
        yield palette


def get_random_palette() -> List[str]:
    """
    ### Get Random Palette
    Returns a random color palette.
    """
    return choice(_corpus_colors.palettes)


def get_material_color(color: _corpus_colors.MaterialColor) -> Dict:
    """
    ### Get Material Colors
    Returns a `dict` of a material `color` (all of its shades).

    #### Material Colors:
    Google's Material Colors include `red`, `pink`, `purple`, `deeppurple`, `indigo`,
    `blue`, `lightblue`, `cyan`, `teal`, `green`, `lightgreen`, `lime`, `yellow`,
    `amber`, `orange`, `deeporange`, `brown`, `grey`, `bluegrey`.

    Shades of each color: `50`, `100`, `200`, `300`, `400`, `500`, `600`, `700`, `800`,
    `900`, `a100`, `a200`, `a400`, `a700`.
    ```
    #### dict Structure:
    {
        "50": "#ffebee",
        "100": "#ffcdd2",
        ...,
        "a700": "#d50000"
    }
    ```
    """
    return _corpus_colors.material_colors.get(color, None)



""" 
## Ansy
A Python Package for Text Styling/Formatting in the terminal using
ANSI escape sequences.

"""

from __future__ import annotations

from . import exceptions
from . import colors
from .ansy import (
    make_ansi,
    colored,
    colored_gradient,
    make_gradient,
    printc,
    colored_ansy,
    create_style,
    contains_ansi,
    colored_random,
    get_random_color,
    create_random_palette,
    print_all_colors,
    get_all_colors,
    search_colors,
    de_ansi,
    colorname_to_code,
    code_to_colorname,
    hex_to_rgb,
    rgb_to_hex,
    is_valid_hex,
    is_valid_rgb,
    is_valid_attr,
    is_valid_color,
    COLORS_256,
    FGCOLORS_STANDARD,
    BGCOLORS_STANDARD,
    ANSI_REGEX,
    StandardColor,
    Color256,
    Attribute,
    RGBTuple,
    Color,
    ColorMode,
    Quality,
    ANSI_CODES,
    STANDARD_COLORNAMES,
    COLORS256_COLORNAMES,
    ATTRIBUTES,
)

__all__ = [
    "ANSI_CODES",
    "ANSI_REGEX",
    "Attribute",
    "ATTRIBUTES",
    "BGCOLORS_STANDARD",
    "COLORS_256",
    "COLORS256_COLORNAMES",
    "Color",
    "ColorMode",
    "Color256",
    "FGCOLORS_STANDARD",
    "RGBTuple",
    "StandardColor",
    "code_to_colorname",
    "colored",
    "colored_ansy",
    "colored_gradient",
    "colored_random",
    "colorname_to_code",
    "contains_ansi",
    "create_random_palette",
    "create_style",
    "de_ansi",
    "get_all_colors",
    "get_random_color",
    "hex_to_rgb",
    "is_valid_attr",
    "is_valid_color",
    "is_valid_hex",
    "is_valid_rgb",
    "make_ansi",
    "make_gradient",
    "print_all_colors",
    "printc",
    "Quality",
    "rgb_to_hex",
    "search_colors",
    "STANDARD_COLORNAMES",
]



import ansy
from argparse import ArgumentParser


def main():
    args = get_args()

    text = args.text
    fgcolor = args.fgcolor
    bgcolor = args.bgcolor
    color_mode = args.mode

    if color_mode and color_mode not in (4, 8, 24):
        error(f"Invalid color mode: {color_mode}, must be 4, 8, or 24.")

    if fgcolor:
        if not ansy.is_valid_color(fgcolor, color_mode):
            error(f"Invalid foreground color: {fgcolor}")
    else:
        fgcolor = ansy.get_random_color(color_mode)

    if bgcolor and not ansy.is_valid_color(bgcolor, color_mode):
        error(f"Invalid background color: {bgcolor}")

    ansy.printc(text, fgcolor, bgcolor, color_mode=color_mode)


def get_args():
    parser = ArgumentParser(
        usage="ansy [OPTIONS]",
        description="Command-line utility to color text.",
        epilog="ansy v1.1.3",
    )

    parser.add_argument(
        "-t", "--text", help="the text to color", metavar="", default="Welcome to ansy."
    )
    parser.add_argument(
        "-fg", "--fgcolor", help="the foreground color of text", metavar=""
    )
    parser.add_argument(
        "-bg", "--bgcolor", help="the background color of text", metavar=""
    )
    parser.add_argument(
        "-m",
        "--mode",
        help="the color mode (4, 8, or 24)",
        metavar="",
        type=int,
        default=4,
    )

    # Parse and return Args
    return parser.parse_args()


def error(message):
    ansy.printc(f"Error: {message}", fgcolor="light_red")
    exit(1)



# Copyright (c) 2024 Anas Shakeel

from __future__ import annotations
from sys import stdout
from os import isatty, environ
from re import IGNORECASE, compile as re_compile
from io import UnsupportedOperation
from random import choice, randint
from typing import Literal, Iterable, Union

from ._others import (
    StandardColor,
    Attribute,
    Color256,
    RGBTuple,
    Color,
    ColorMode,
    Quality,
    BGCOLORS_STANDARD,
    FGCOLORS_STANDARD,
    ATTRIBUTES,
    COLORS256_COLORNAMES,
    STANDARD_COLORNAMES,
    ANSI_CODES,
    COLORS_256,
)
from .exceptions import (
    InvalidColorError,
    RGBError,
    HexError,
    StyleError,
    ColorModeError,
)


# Compiled REGEX Patterns
ANSI_REGEX = re_compile(
    r"[\u001B\u009B][\[\]()#;?]*(?:(?:(?:(?:;[-a-zA-Z\d\/#&.:=?%@~_]+)*|[a-zA-Z\d]+(?:;[-a-zA-Z\d\/#&.:=?%@~_]*)*)?\u0007)|(?:(?:\d{1,4}(?:;\d{0,4})*)?[\dA-PR-TZcf-nq-uy=><~]))"
)
HEX_REGEX = re_compile(r"^#?((?:[0-9a-fA-F]{3}){1,2})$", IGNORECASE)
ANSY_STR_REGEX = re_compile(r"(@(?:\w|\d)+\[.*?\])", IGNORECASE)
# ANSY_STR_REGEX with capturing groups
ANSY_STR_CAPT_REGEX = re_compile(r"@((?:\w|\d)+)\[(.*?)\]", IGNORECASE)


# This function is borrowed from 'termcolor' library. licensed under the MIT License.
# Full license text can be found in the THIRD_PARTY_LICENSE file.
def _can_do_colour(*, no_color: bool = None, force_color: bool = None) -> bool:
    """
    Check env vars and for tty/dumb terminal.

    #### First check overrides:
    User-level configuration files and per-instance command-line arguments should
    override `$NO_COLOR`. A user should be able to export `$NO_COLOR` in their shell
    configuration file as a default, but configure a specific program in its
    configuration file to specifically enable color.

    https://no-color.org
    """
    if no_color is not None and no_color:
        return False
    if force_color is not None and force_color:
        return True

    # Then check env vars:
    if "ANSI_COLORS_DISABLED" in environ:
        return False
    if "NO_COLOR" in environ:
        return False
    if "FORCE_COLOR" in environ:
        return True

    # Then check system:
    if environ.get("TERM") == "dumb":
        return False
    if not hasattr(stdout, "fileno"):
        return False

    try:
        return isatty(stdout.fileno())
    except UnsupportedOperation:
        return stdout.isatty()


def colored(
    text: str,
    fgcolor: Color = None,
    bgcolor: Color = None,
    attrs: Iterable[Attribute] = None,
    color_mode: ColorMode = 4,
    *,
    no_color: bool = None,
    force_color: bool = None,
) -> str:
    """
    ### Colored
    Colorize text using `4`, `8`, or `24` bit colors. If `fgcolor`, `bgcolor`, and `attrs` all are `None`,
    the `text` itself is returned without any ansi codes added.

    #### ARGS:
    - `text`: the text to colorize
    - `fgcolor`: the foreground color
    - `bgcolor`: the background color
    - `attrs`: the attributes to apply
    - `color_mode`: the colormode to use (defualt is `4`)

    #### Acceptable `4-bit` colors (Foreground|Background):
        `black`, `red`, `green`, `yellow`, `blue`, `magenta`, `cyan`, `white`,
        `light_grey`, `dark_grey`, `light_red`, `light_green`, `light_yellow`,
        `light_blue`, `light_magenta`, `light_cyan`.

    #### Acceptable `8-bit` colors (Foreground|Background):
        - `ansy.print_all_colors()` to print all 8-bit colors.
        - `ansy.get_all_colors()` to iterate through all 8-bit colors.

    #### Acceptable `24-bit` colors (Foreground|Background):
        an RGB tuple or Hex color code

    #### Available attributes:
        `bold`, `dark`, `underline`, `blink`, `reverse`, `concealed`,
        `double-underline`, `overline`, `strike`, `italic`

    ```
    # Example
    >> text = 'Hello, ansy!'
    >> colored(text, 'red', 'black', ['bold', 'blink'])
    >>
    >> colored(text, 'plum', color_mode=8)  # 8-bit color: name
    >> colored(text, 215, color_mode=8) # 8-bit color: code
    >>
    >> colored(text, '#B00B1E', color_mode=24) # 24-bit color: Hex
    >> colored(text, (255,100,100), color_mode=24) # 24-bit color: RGB
    ```

    Raises `ColorModeError` if:
    - `color_mode` is invalid

    Raises `InvalidColorError` if:
    - `fgcolor` is invalid
    - `bgcolor` is invalid

    Raises `AttributeError` if:
    - `attrs` contains an invalid attribute

    """
    # Boolean, None, or Empty string would return ""
    if isinstance(text, bool) or text is None or text == "":
        return ""

    result = str(text)

    if not _can_do_colour(no_color=no_color, force_color=force_color):
        return result

    # If no styling requested, return result
    if fgcolor == None and bgcolor == None and attrs == None:
        return result

    return make_ansi(fgcolor, bgcolor, attrs, color_mode) + result + ANSI_CODES["reset"]


def printc(
    text: str,
    fgcolor: Color = None,
    bgcolor: Color = None,
    attrs: Iterable[Attribute] = None,
    color_mode: ColorMode = 4,
    *,
    sep=" ",
    end="\n",
    file=None,
    flush=False,
    no_color: bool = None,
    force_color: bool = None,
):
    """
    ### Print Colored
    Prints whatever returns from `ansy.colored`.

    ```
    # Example
    >> ansy.printc('Print this.', 'light_red', attrs=['bold'])
    ```
    """
    print(
        colored(
            text,
            fgcolor,
            bgcolor,
            attrs,
            color_mode,
            no_color=no_color,
            force_color=force_color,
        ),
        sep=sep,
        end=end,
        file=file,
        flush=flush,
    )


def colored_ansy(
    text: str, style: dict, *, no_color: bool = None, force_color: bool = None
) -> str:
    """
    ### Colored Ansy
    Adds the styling defined in `style` to wherever an `ansy string`
    is found in the `text`.

    #### ARGS:
    - `text`: the text containing the `ansy strings`
    - `style`: a dictionary containing the style(s).

    ##### Style dict:
    `style` is a dictionary that defines the styles to apply to these
    `ansy_strings`. Each style in the `style` dictionary is itself a dictionary.
    Purpose of this structure is so that you could define all your styles in one
    dictionary and then use it everywhere, instead of creating a dictioanry for
    each style and create a mess in your code.

    ```
    # Example Usage
    >> style = {
    ..     "my_style": {
    ..         "color_mode": 24,
    ..         "fgcolor": (200,100,50),
    ..         "bgcolor": '#121212',
    ..         "attrs": ["bold", "reverse"]
    ..     }
    .. }
    >>
    >> # Style can also be created by create_style()
    >> style['new_style'] = ansy.create_style(24, 'fff', (200,0,0), attrs=['bold'])
    >>
    >> ansy.colored_ansy("This is an @new_style[ansy string]", style)
    'This is an ansy string'
    >> # above string is formatted as defined by "new_style"
    ```
    `style` dictionary can have as many styles as you want, just make sure you use
    the same name for the style in `ansy string` as defined in `style`

    Raises `StyleError` if:
    - style used in `ansy string` not found in `style` dict

    Raises `ColorModeError` if:
    - `color_mode` provided in style, is invalid

    """
    # Find all ansy_strings and split them
    matches: list = ANSY_STR_REGEX.split(text)
    if not matches:
        return None

    parsed = [""] * len(matches)
    for index, m in enumerate(matches):
        # Parse, format and add into the parsed:list
        parsed[index] = _parse_ansy(
            m, style, no_color=no_color, force_color=force_color
        )

    return "".join(parsed)


def _parse_ansy(
    ansy: str, style: dict, *, no_color: bool = None, force_color: bool = None
) -> str:
    """
    Parse `ansy` string and return the formatted ansy.

    #### ARGS:
    - `ansy`: the ansy string
    - `style`: the styles dictionary which contains the styles (formats if style found, raises `ValueError` otherwise)
    """
    if ansy_matches := ANSY_STR_CAPT_REGEX.match(ansy):
        # Parse ansy_matches and apply formatting
        if ansy_matches.group(1) not in style:
            err = f"Undefined style: '{ansy_matches.group(1)}'"
            raise StyleError(err)

        # Style dict for currently matched ansy_string
        current_style: dict = style[ansy_matches.group(1)]

        # Colorize the ansy strings
        return colored(
            ansy_matches.group(2),
            current_style["fgcolor"],
            current_style["bgcolor"],
            current_style["attrs"],
            current_style["color_mode"],
            no_color=no_color,
            force_color=force_color,
        )

    # Returns the string if not ansy
    return ansy


def create_style(
    color_mode: ColorMode = 4,
    fgcolor: Color = None,
    bgcolor: Color = None,
    attrs: Iterable[Attribute] = None,
) -> dict:
    """
    ### Create Style
    Creates a style `dict` to use in `colored_ansy()` as style.

    Sole purpose of this function is to make the process creating a style easier
    and safer. it validates the `fgcolor`, `bgcolor` and `attrs` to make
    sure they are valid.

    #### ARGS:
    - `color_mode`: the color mode (`4-bit`, `8-bit`, `24-bit`)
        - `4` means use 4-bit colors (`16` Standard colors)
        - `8` means use 8-bit colors (`256` colors)
        - `24` means use 24-bit colors (approx. `16.7 million` colors)
    - `fgcolor`: the foreground color
    - `bgcolor`: the background color
    - `attrs`: the attributes list

    ```
    >> # Example Usage
    >> fg = (255,100,100)
    >> bg = (50,0,50)
    >> attrs = ["bold", "underline"]
    >> styles['my_style'] = ansy.create_style(24, fg, bg, attrs)
    >> styles
    {
        "my_style": {
            "color_mode": 24,
            "fgcolor": (255,100,100),
            "bgcolor": (50,0,50),
            "attrs": ["bold", "underline"]
        }
    }
    ```

    Raises `InvalidColorError` if:
    - `fgcolor` is not a valid color
    - `bgcolor` is not a valid color

    Raises `ColorModeError` if:
    - `color_mode` is invalid

    Raises `AttributeError` if:
    - `attrs` includes an invalid attribute string

    """
    # Invalid color_mode ?
    if not _is_valid_colormode(color_mode):
        raise ColorModeError(f"Invalid color_mode: {color_mode}")

    # Color Validation
    fgcolor, bgcolor = _validate_colors((fgcolor, bgcolor), color_mode)

    # Attributes Valiadation
    attrs = _validate_attrs(attrs)

    # Create the dictionary
    return {
        "color_mode": color_mode,
        "fgcolor": fgcolor,
        "bgcolor": bgcolor,
        "attrs": attrs,
    }


def colored_gradient(
    text: str,
    start_color: Union[RGBTuple, str],
    end_color: Union[RGBTuple, str],
    quality: Quality = "medium",
    reverse: bool = False,
) -> str:
    """
    ### Colored Gradient (24-bit)
    Apply horizontal gradient on `text`.

    ##### NOTE:
    This method (when `quality` is high) assigns a color (ANSI sequence for RGB colors) to each character of text.
    an ANSI Sequence for one RGB color can be from `17` to `23` characters long.

    It goes like: `len(text) * 17 or 23`.

    So if input `text` is 10 characters long, output text will be
    approx. `170` or `230` characters long!

    An input of `64 characters` outputs approx. `1,472 characters`.
    ##### Point being: Use higher quality only when necessary!

    #### ARGS:
    - `text`: the text to apply the gradient to
    - `start_color`: starting foreground color (`RGBTuple` or `Hex code`)
    - `end_color`: ending foreground color (`RGBTuple` or `Hex code`)
    - `quality`: quality of the gradient (choose from `high`, `medium`, or `low`)
    - `reverse`: reverses the gradient to be (right to left) instead.

    Raises `InvalidColorError` if:
    - `start_color` is an invalid `24-bit` color
    - `end_color` is an invalid `24-bit` color

    Raises `AssertionError` if:
    - `quality` is not set to `high` or `medium` or `low`
    - `start_color` or `end_color` are `None`
    """
    if not text:
        return ""

    assert (
        start_color != None or end_color != None
    ), "start_color and end_color must not be None"

    # Colors Validation
    start_color, end_color = _validate_colors((start_color, end_color), 24)

    # Convert to RGBs, if Hexcodes
    start_color = start_color if is_valid_rgb(start_color) else hex_to_rgb(start_color)
    end_color = end_color if is_valid_rgb(end_color) else hex_to_rgb(end_color)

    assert quality in ("high", "medium", "low")

    # Length of text
    text_length = len(text)

    # Create the gradient (Generator not actual values)
    gradient = make_gradient(start_color, end_color, text_length, reverse=reverse)

    batch_size = 1
    if quality == "low":
        batch_size = len(text) // 4
    elif quality == "medium":
        batch_size = len(text) // 8
    else:
        # High quality, already set
        pass

    # Clip batch size
    batch_size = max(batch_size, 1)  # to 1, if less...
    batch_size = min(batch_size, text_length)  # to text_length, if more...

    # Apply the gradient
    return _apply_gradient(text, gradient, batch_size)


def _apply_gradient(text: str, gradient: list, batch_size: int) -> str:
    """Apply `gradient` to `text`"""
    fmt_str = ANSI_CODES["24bit"]
    gradient_text = ""
    read_chars = 0
    batch = []

    # Iterate through the text and gradient
    for char, color in zip(text, gradient):
        batch.append(char)  # Add char to batch
        read_chars += 1

        # Is batch full?
        if len(batch) == batch_size:
            gradient_text += fmt_str % (38, *color, "".join(batch))
            batch = []  # Start new batch
            continue

        # Is this last batch?
        if read_chars >= len(text):
            gradient_text += fmt_str % (38, *color, "".join(batch))

    return gradient_text + ANSI_CODES["reset"]


def make_gradient(
    start_color: RGBTuple, end_color: RGBTuple, steps: int = 10, reverse: bool = False
):
    """
    ### Make Gradient (24-bit)
    Make a gradient between two RGB colors `start_color` and `end_color` by
    linearly interpolating between these colors for a given number of `steps`.
    Returns a `generator object`.

    #### ARGS:
    - `start_color`: starting rgb color
    - `end_color`: ending rgb color
    - `steps`: steps to interpolate across (minimum `2`)
    - `reverse`: flip the gradient horizontally

    ```
    # Example
    >> start_color = (240, 194, 123)
    >> end_color = (255, 85, 85)
    >> gradient = ansy.make_gradient(start_color, end_color, steps=4)
    >>
    >> gradient
    <generator object make_gradient at 0x0000018F0B580700>
    >>
    >> list(gradient)
    [(240, 194, 123), (245, 157, 110), (250, 121, 97), (255, 85, 85)]
    ```

    Raises `InvalidColorError` if:
    - `start_color` is not a valid RGBTuple
    - `end_color` is not a valid RGBTuple
    """
    # Validate the RGBs
    if start_color and not is_valid_rgb(start_color):
        raise InvalidColorError(f"Invalid RGB: {start_color}")
    if end_color and not is_valid_rgb(end_color):
        raise InvalidColorError(f"Invalid RGB: {end_color}")

    # Clip steps to 2 if lesser.
    steps = 2 if steps < 2 else steps

    # Flip the colors, if asked...
    if reverse:
        start_color, end_color = end_color, start_color

    # Generate the gradient colors
    for step in range(steps):
        r = int(start_color[0] + (end_color[0] - start_color[0]) * (step / (steps - 1)))
        g = int(start_color[1] + (end_color[1] - start_color[1]) * (step / (steps - 1)))
        b = int(start_color[2] + (end_color[2] - start_color[2]) * (step / (steps - 1)))

        yield (r, g, b)


def colored_random(
    text: str,
    target: Literal["all", "words", "chars"] = "all",
    color_mode: ColorMode = 4,
    custom_palette: Iterable = None,
    attrs: Iterable[Attribute] = None,
    random_attrs: bool = False,
) -> str:
    """
    ### Colored Random
    Apply a random color to `text`. colors are chosen from a colorsystem specified by `color_mode`.

    #### ARGS:
    - `text`: the text string
    - `target`: apply random colors to `chars`, `words` or `all` string.
    - `color_mode`: the colormode to use
    - `custom_palette`: iterable of user-defined colors (default is `None` which uses random colors from the whole range of colors)
    - `attrs`: the attributes to apply (attributes are NOT randomized by default)
    - `random_attrs`: whether to apply attributes randomly or not

    ```
    # Example
    >> text = 'This is just a string.'
    >> ansy.colored_random(text, 'all', 4)
    'This is just a string.' # This is randomly colored
    >>
    >> text = 'This is just a string.'
    >> ansy.colored_random(text, 'words', 4)
    'This is just a string.'
    >>
    >> palette = ['light_red', 'blue']
    >> ansy.colored_random(text, 'words', 4, palette)
    'This is just a string.'
    >>
    >> # Not recommended to use 'chars' with 24bit colors, if string is longer than 50 characters
    >> ansy.colored_random(text, 'chars', 24)
    'This is just a string.'
    >>
    >> # Why not you might ask! let me show you
    >> f = ansy.colored_random('This', 'chars', 24)
    >> print([f])
    ['\x1b[38;2;5;155;190mT\x1b[0m\x1b[38;2;119;11;99mh\x1b[0m\x1b[38;2;24;50;220mi\x1b[0m\x1b[38;2;178;45;244ms\x1b[0m']
    >>
    >> # for a string of 4 characters, it outputs 113 characters! and for what? just to see some random colors in the terminal?
    ```

    Raises `ValueError` if:
    - `target` is not set to `'chars'`, `'words'`, `'all'`

    Raises `ColorModeError` if:
    - `color_mode` is invalid

    Raises `InvalidColorError` if:
    - `custom_palette` contains an invalid color

    Raises `AttributeError` if:
    - `attrs` contains an invalid attribute

    """
    if not text:
        return ""

    # Palette Validation
    if custom_palette:
        _validate_colors(custom_palette, color_mode)

    # Attrs validation
    attrs = _validate_attrs(attrs)

    # Break the text
    if target == "chars":
        broken_text: list = list(text)  # Into letters
    elif target == "words":
        broken_text: list = text.split(" ")  # Into words
    elif target == "all":
        broken_text: list = [text]  # Whole text
    else:
        # Raise error
        raise ValueError("target must be 'chars', 'words' or 'all'")

    colorized_items: list = [""] * len(broken_text)
    for i, chunk in enumerate(broken_text):
        colorized_items[i] = _apply_random_colors(
            chunk, color_mode, custom_palette, attrs, random_attrs
        )
    # Join separator
    sep = ""
    if target == "words":
        sep = " "

    return sep.join(colorized_items)


def _apply_random_colors(
    text: str,
    color_mode: ColorMode,
    custom_palette: Iterable,
    attrs: Iterable[Attribute],
    random_attrs: bool,
) -> str:
    """Apply random colors and/or attributes"""
    # Choose a random color or from palette
    if custom_palette is not None:
        color = choice(custom_palette)
    else:
        color = get_random_color(color_mode)

    # Choose random attribute
    new_attrs = [choice(attrs)] if (attrs and random_attrs) else attrs

    # Format current item (word or char)
    return colored(text, color, color_mode=color_mode, attrs=new_attrs)


def get_random_color(color_mode: ColorMode = 4) -> Union[str, RGBTuple]:
    """
    ### Get Random Color
    Returns a random color from a colorsystem specified by `color_mode`.

    #### Values Returned for each colormode:
    - `4`: Returns the name of a 4-bit color
    - `8`: Returns the name of an 8-bit color
    - `24`: Returns the `RGBTuple`

    ```
    # Example
    >> ansy.get_random_color(4)
    'light_green'
    >> ansy.get_random_color(8)
    'plum'
    >> ansy.get_random_color(24)
    (184, 30, 57)
    ```

    Raises `ColorModeError` if:
    - `color_mode` is invalid.

    """
    if color_mode == 4:
        # Choose a random 4bit color
        return choice(STANDARD_COLORNAMES)
    elif color_mode == 8:
        # Choose a random 8bit color
        return choice(COLORS256_COLORNAMES)
    elif color_mode == 24:
        # Create a random RGBTuple
        return (randint(0, 255), randint(0, 255), randint(0, 255))
    else:
        # Raise error
        raise ColorModeError(f"Invalid color_mode: {color_mode}")


def create_random_palette(color_mode: ColorMode, n: int) -> list:
    """
    ### Create Palette
    Returns a list of colors from `color_mode`-bit color system.

    #### ARGS:
    - `color_mode`: the colormode to use
    - `n`: number of colors to include in palette

    Raises `TypeError` if:
    - `n`: is not an integer
    """
    if type(n) != int:
        raise TypeError("n is expected to be an integer.")
    return [get_random_color(color_mode) for _ in range(n)]


def de_ansi(text: str) -> str:
    """
    ### De-Ansi
    Removes all ansi codes from `text`. It uses `ANSI_REGEX` pattern.
    Returns the `text` as is, If no ANSI sequences found.

    #### ARGS:
    - `text`: the text to de_ansi-fy

    ```
    # Example
    >> # Ansi containting string
    >> ansi_string = colored("This is clean.", "red", "white", ["bold", "dark"])
    >>
    >> de_ansi(ansi_string)
    'This is clean.'
    ```

    Raises `TypeError` if:
    - `text` is not a `str` object
    """
    if not isinstance(text, str):
        raise TypeError(f"'text' is expected to be str: not {type(text)}")

    # Recognize all ANSI Patterns in a string
    return ANSI_REGEX.sub("", text)


def contains_ansi(text: str) -> bool:
    """
    ### Contains ansi
    Returns `True` if text contains ansi codes, else `False`.

    #### Example:
    ```
    >> text = colored("Formatted String", fgcolor="red")
    >> contains_ansi(text)
    True
    >> text = de_ansi(text)
    >> contains_ansi(text)
    False
    ```
    """
    return True if ANSI_REGEX.search(text) else False


def print_all_colors():
    """
    ### Print All Colors
    Print all 256 colors with their name and code.
    (sorted by codes in asc. order)
    """
    for code, clr in sorted(zip(COLORS_256.values(), COLORS_256.keys())):
        print(
            " ",
            colored("    ", bgcolor=clr, color_mode=8),
            f"{colored(code, code, color_mode=8)}: {clr}",
        )


def get_all_colors(sort_by_name: bool = False):
    """
    ### Get All Colors
    Yields a `tuple` of `color_name` and it's `code`. (sorted by `code`)

    These colors include `16` Standard Colors, `216` Extended
    Colors, and Remaining `24` Colors are Greyscale.

    #### ARGS:
    - `sort_by_name`: Sorts colors by name (default is `False` which sorts by code)

    ```
    # Example Usage
    >> from ansy import get_all_colors
    >> for color, code in get_all_colors():
    >>     print(f"{color}, {code}")
    aquamarine, 79
    aquamarine_1, 86
    --snip--
    yellow_pale, 187
    ```
    """
    zipped = zip(COLORS_256.values(), COLORS_256.keys())
    if sort_by_name:
        zipped = zip(COLORS_256.keys(), COLORS_256.values())

    for c1, c2 in sorted(zipped):
        if sort_by_name:
            yield c1, c2
        else:
            yield c2, c1


def search_colors(query: str):
    """
    ### Search [8-bit] Colors
    Yields all colors (`name`, `code`) that contain `query` in their name.

    #### ARGS:
    - `query`: the string to look for...
    """
    if not query:
        return None

    for clr, code in get_all_colors(True):
        if query in clr:
            yield clr, code


def make_ansi(
    fgcolor: Color = None,
    bgcolor: Color = None,
    attrs: Iterable[Attribute] = None,
    color_mode=4,
) -> str:
    """
    ### Make Ansi
    Make an `Ansi Escape Sequence` from the args given.
    Returns the sequence without `Ansi Reset Code`.

    #### ARGS:
    - `fgcolor`: the foreground color (any color that belongs to `color_mode`)
    - `bgcolor`: the background color (any color that belongs to `color_mode`)
    - `attrs`: the attributes
    - `color_mode`: the color mode to use

    ```
    # Example
    >> ansy.make_ansi("light_green", "dark_grey", color_mode=4)
    >> ansy.make_ansi(19, "plum", attrs=['bold', 'italic'], color_mode=8)
    >> ansy.make_ansi("#A22", (50,50,100), color_mode=24)
    ```

    Raises `InvalidColorError` if:
    - `fgcolor` is invalid according to color_mode
    - `bgcolor` is invalid according to color_mode

    Raises `ColorModeError` if:
    - `color_mode` is invalid
    """
    if not _is_valid_colormode(color_mode):
        raise ColorModeError(f"Invalid color_mode: {color_mode}")

    # Color validation
    fgcolor, bgcolor = _validate_colors((fgcolor, bgcolor), color_mode)

    # Attrs validation
    attrs = _validate_attrs(attrs)

    # Normalizing the colors to be used in ANSI codes
    if fgcolor:
        fgcolor = _normalize_color(fgcolor, color_mode)
    if bgcolor:
        bgcolor = _normalize_color(bgcolor, color_mode, True)

    result = ""
    fmt_str = ANSI_CODES[f"{color_mode}bit"]

    # Adding Colors
    if color_mode == 4:
        if fgcolor:
            result = fmt_str % (fgcolor, result)

        if bgcolor:
            result = fmt_str % (bgcolor, result)

    elif color_mode == 8:
        if fgcolor:
            result = fmt_str % (38, fgcolor, result)

        if bgcolor:
            result = fmt_str % (48, bgcolor, result)

    else:  # 24-bit
        if fgcolor:
            result = fmt_str % (38, *fgcolor, result)

        if bgcolor:
            result = fmt_str % (48, *bgcolor, result)

    # Adding Attributes
    if attrs:
        for a in attrs:
            result = ANSI_CODES["attr"] % (ATTRIBUTES[a], result)

    return result


def _normalize_color(color: Color, color_mode: ColorMode, bg: bool = False) -> Color:
    """
    Normalize `color` to be used in ansi escape sequences.
    Normalized colors are safe to use in ansi sequences
    without any additional validations.

    #### NOTE:
    Caller is responsible for validating the colors.
    This method throws errors otherwise.

    #### ARGS:
    - `color` : the color to normalize
    - `color_mode` : the colormode to use
    - `bg` : normalize as `BG` color? only applicable on 4-bit color_mode
    """

    if color_mode == 4:
        return BGCOLORS_STANDARD[color] if bg == True else FGCOLORS_STANDARD[color]

    elif color_mode == 8:
        return colorname_to_code(color)

    elif color_mode == 24:
        return hex_to_rgb(color) if is_valid_hex(color) else color
    else:
        return None


def _validate_colors(colors: tuple, color_mode: ColorMode) -> tuple:
    """
    ### Validate Colors
    Validate the colors based on `color_mode`. returns the `colors` if all good.

    #### Syntax for `colors`:
    - `(fgcolor, bgcolor, someothercolor, etc_color)`

    ```
    # Example
    >> fgcolor, bgcolor = 'light_red', None
    >> _validate_colors((fgcolor, bgcolor), color_mode=4)
    ('light_red', None)
    >>
    >> _validate_colors(('light_red', 'invalid_color'), color_mode=8)
    InvalidColorError: Invalid 8-bit color: invalid_color
    ```

    Raises `InvalidColorError` if:
    - `color` is invalid
    """
    for color in colors:
        if color != None and not is_valid_color(color, color_mode):
            err = f"Invalid {color_mode}-bit color: {color}"
            raise InvalidColorError(err)

    return colors


def _validate_attrs(attrs: Iterable[Attribute]) -> list:
    """
    ### Validate Attrs
    Validate the `attrs`. returns the `attrs` if all good.

    ```
    # Example
    >> _validate_attrs(['bold', 'italic'])
    ['bold', 'italic']
    >>
    >> _validate_attrs(['underline', 'invalid_attr'])
    AttributeError: Invalid attribute: invalid_attr
    ```

    Raises `AttributeError` if:
    - `attrs` contais an invalid attribute
    """
    if attrs:
        for attr in attrs:
            if not is_valid_attr(attr):
                err = f"Invalid attribute: {attr}"
                raise AttributeError(err)

    return attrs


def _is_valid_colormode(color_mode: ColorMode) -> bool:
    """Validates colormode, Returns `True` if valid, `False` otherwise"""
    if color_mode in (4, 8, 24):
        return True
    return False


def _is_color_8bit(color: Union[int, str]) -> bool:
    """
    Returns `True` if `clr` is a recognized color, else `False`.

    - `clr`: a color name or code (from 256 color system)
    """
    if not isinstance(color, (int, str)):
        return False

    if color in COLORS_256 or color in COLORS_256.values():
        # If color name OR color code
        return True

    return False


def is_valid_color(color: Color, color_mode: ColorMode) -> bool:
    """
    ### Is Valid Color?
    validates the color based on `color_mode`. Returns `True` if color
    is valid, `False` for otherwise.

    #### ARGS:
    - `color`: the color to validate
    - `color_mode`: the color mode of `color` (can be `4`, `8`, `24`)
    """
    if color_mode == 4:
        return color in STANDARD_COLORNAMES
    elif color_mode == 8:
        return _is_color_8bit(color)
    elif color_mode == 24:
        return is_valid_rgb(color) or is_valid_hex(color)
    else:
        return False


def is_valid_rgb(rgb: RGBTuple) -> bool:
    """
    ### Is valid rgb
    Returns `True` if `rgb` is valid rgb, else `False`.
    `rgb` must be a `tuple` not `list`, not `set` not anything else.

    #### Valid RGB:
    RGB tuple is valid if:
    - There are three integer values.
    - `R` is an int from `[0-255]`
    - `G` is an int from `[0-255]`
    - `B` is an int from `[0-255]`

    #### ARGS:
    - `rgb`: the rgb tuple to validate

    """
    # Must be a tuple
    if not isinstance(rgb, tuple):
        return False

    # Must have 3 values
    if len(rgb) != 3:
        return False

    # Must lie in range 0-255
    for v in rgb:
        if type(v) != int or v > 255 or v < 0:
            return False

    return True


def is_valid_hex(hexcode: str) -> bool:
    """
    ### Is valid hex
    Returns `True` if `hexcode` is valid hex color code, else `False`.

    #### ARGS:
    - `hexcode`: the hex color code to validate
    """
    # Must be a str
    if type(hexcode) != str:
        return False

    if HEX_REGEX.match(hexcode):
        return True

    return False


def is_valid_attr(attr: Attribute) -> bool:
    """
    ### Is valid attr
    Returns `True` if `attr` is a valid Attribute, else `False`.

    #### ARGS:
    - `attr`: the attribute string

    #### Valid Attributes:
    `bold`, `dark`, `underline`, `blink`, `reverse`, `concealed`,
    `double-underline`, `overline`, `strike`, `italic`

    ```
    # Example
    >> is_valid_attr('bold')
    True
    >> is_valid_attr('ansy')
    False
    ```
    """
    if attr in ATTRIBUTES:
        return True

    return False


def colorname_to_code(color: Color256) -> Union[int, None]:
    """
    ### Colorname to Code
    Returns the `code` for the `color`.
    Returns `None` if color is not a valid 8-bit color.
    If `color` is already an 8-bit color code, returns the code.

    #### ARGS:
    - `color`: name of a color (from 256 color system)

    ```
    # Example
    >> colorname_to_code("red")
    1
    >> colorname_to_code("red-ish")
    None
    >> colorname_to_code(1)
    1
    >> colorname_to_code(300)
    None
    ```

    #### FunFact:
    This method raises no errors whatsoever!
    """
    if not is_valid_color(color, 8):
        return None

    return COLORS_256.get(color, color)


def code_to_colorname(color: int) -> Union[str, None]:
    """
    ### Code to Colorname
    Returns the color name for the `color` code.
    Returns `None` if color is not a valid color. If `color` is already a valid
    colorname, Returns the name.

    #### ARGS:
    - `color`: code of a color (from 256 color system `i.e. [0-255]`)

    ```
    # Example
    >> code_to_colorname(1)
    'red'
    >> code_to_colorname(-2)
    None
    >> code_to_colorname("blue")
    'blue'
    ```

    #### FunFact:
    This method raises no errors whatsoever!

    """
    # Ensure color is valid
    if not is_valid_color(color, 8):
        return None

    # Search colorcode, return if found
    for clr, c in COLORS_256.items():
        if c == color:
            return clr

    # At this point, its a valid color but not a code,
    # means its already a colorname
    return color


def hex_to_rgb(hexcode: str) -> RGBTuple:
    """
    ### Hex to RGB
    Converts a `hex` color code into an `RGB tuple`.
    It is case-insensitive and also recognizes 3-lettered hex
    color codes `e.g #FFF, #9ca etc.`

    #### Valid HEX Color Code:
    - `#` symbol is optional
    - length must either be `3` or `6` (excluding `#` symbol)
    - each letter must only be 0-9, a-f
    - case-insensitive

    #### ARGS:
    - `hexcode`: a hex color code string

    ```
    # Example
    >> ansy.hex_to_rgb("#FFFFFF")
    (255, 255, 255)
    >> ansy.hex_to_rgb("Fff")
    (255, 255, 255)
    ```

    Raises `HexError` if:
    - `hexcode` is not a valid hex color code
    """
    if not is_valid_hex(hexcode):
        raise HexError(f"Invalid hexcode: {hexcode}")

    new_hex = hexcode.lstrip("#")  # Strip the hash

    # Convert to 6-letter hex (if not already)
    if len(new_hex) == 3:
        new_hex = "".join([l + l for l in new_hex])

    # Convert the hex to RGB
    return tuple(int(new_hex[i : i + 2], 16) for i in (0, 2, 4))


def rgb_to_hex(rgb: RGBTuple, with_symbol: bool = True) -> str:
    """
    ### RGB to Hex
    Converts an `rgb` tuple into Hex color code.

    #### ARGS:
    - `rgb`: RGB tuple to convert to hex
    - `with_symbol`: Whether to include the `#` symbol in output

    ```
    # Example
    >> ansy.hex_to_rgb((255,255,255), with_symbol=True)
    "#FFFFFF"
    >> ansy.hex_to_rgb((0,0,0), with_symbol=False)
    "000000"
    ```

    Raises `RGBError` if:
    - `rgb` is not a valid RGB tuple
    """
    # Validate the RGB tuple
    if not is_valid_rgb(rgb):
        raise RGBError(f"rgb {rgb} is not a valid RGB tuple.")

    # Conversion...
    hexcode = "#" if with_symbol else ""
    for value in rgb:
        h = format(value, "X")  # Convert to hex
        if len(h) == 1:
            h = "0" + h  # Add zero before the hex if length:1
        hexcode += h

    return hexcode



from __future__ import annotations

from typing import Literal, Tuple, Union


# ANSI Format strings
ANSI_CODES = {
    "reset": "\033[0m",
    "attr": "\033[%dm%s",
    "4bit": "\033[%dm%s",
    "8bit": "\033[%d;5;%dm%s",
    "24bit": "\033[%d;2;%d;%d;%dm%s",
}

# ANSI Attribute codes
ATTRIBUTES: dict[Attribute, int] = {
    "bold": 1,
    "dark": 2,
    "italic": 3,
    "underline": 4,
    "blink": 5,
    "reverse": 7,
    "concealed": 8,
    "strike": 9,
    "double-underline": 21,
    "overline": 53,
}

# Standard Foreground Colors [4-bit]
FGCOLORS_STANDARD: dict[StandardColor, int] = {
    "black": 30,
    "grey": 30,  # Actually black but kept for backwards compatibility
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "light_grey": 37,
    "dark_grey": 90,
    "light_red": 91,
    "light_green": 92,
    "light_yellow": 93,
    "light_blue": 94,
    "light_magenta": 95,
    "light_cyan": 96,
    "white": 97,
}

# Standard Background colors [4-bit]
BGCOLORS_STANDARD: dict[StandardColor, int] = {
    "black": 40,
    "grey": 40,  # Actually black but kept for backwards compatibility
    "red": 41,
    "green": 42,
    "yellow": 43,
    "blue": 44,
    "magenta": 45,
    "cyan": 46,
    "light_grey": 47,
    "dark_grey": 100,
    "light_red": 101,
    "light_green": 102,
    "light_yellow": 103,
    "light_blue": 104,
    "light_magenta": 105,
    "light_cyan": 106,
    "white": 107,
}

# 8-Bit [256 colors]
COLORS_256: dict[Color256, int] = {
    "black": 0,
    "grey": 16,
    "red": 1,
    "green": 2,
    "yellow": 3,
    "blue": 4,
    "magenta": 5,
    "cyan": 6,
    "light_grey": 7,
    "dark_grey": 8,
    "light_red": 9,
    "light_green": 10,
    "light_yellow": 11,
    "light_blue": 12,
    "light_magenta": 13,
    "light_cyan": 14,
    "smokewhite": 15,
    "white": 231,
    "tan": 180,
    "honeydew": 194,
    "light_coral": 210,
    "cornsilk": 230,
    "plum": 183,
    "plum_1": 176,
    "plum_2": 219,
    "plum_dim": 96,
    "plum_dim_1": 139,
    "khaki": 185,
    "khaki_1": 228,
    "khaki_dark": 143,
    "yellow_1": 226,
    "yellow_2": 184,
    "yellow_3": 190,
    "yellow_dark": 100,
    "yellow_pale": 187,
    "yellow_green": 106,
    "yellow_green_1": 148,
    "yellow_green_2": 154,
    "aquamarine": 79,
    "aquamarine_1": 86,
    "aquamarine_2": 122,
    "violet": 177,
    "violet_dark": 92,
    "violet_dark_1": 128,
    "blue_violet": 57,
    "orchid": 170,
    "orchid_medium": 133,
    "orchid_medium_1": 134,
    "orchid_medium_2": 171,
    "orchid_light": 207,
    "orchid_light_1": 212,
    "orchid_light_2": 213,
    "gold": 220,
    "gold_1": 142,
    "gold_2": 178,
    "goldenrod": 179,
    "goldenrod_light": 186,
    "goldenrod_light_1": 221,
    "goldenrod_light_2": 222,
    "goldenrod_light_3": 227,
    "goldenrod_dark": 136,
    "salmon": 209,
    "salmon_dim": 137,
    "salmon_light": 173,
    "salmon_light_1": 216,
    "brown_rosy": 138,
    "brown_sandy": 215,
    "navajo_white": 144,
    "navajo_white_1": 223,
    "misty_rose": 181,
    "misty_rose_1": 224,
    "thistle": 182,
    "thistle_1": 225,
    "wheat": 101,
    "wheat_light": 229,
    "orange": 172,
    "orange_green": 58,
    "orange_light_1": 208,
    "orange_light_2": 214,
    "orange_dim": 94,
    "orange_dim_1": 130,
    "orange_red_1": 166,
    "orange_red_2": 202,
    "magenta_1": 127,
    "magenta_2": 163,
    "magenta_3": 164,
    "magenta_4": 165,
    "magenta_light": 200,
    "magenta_light_1": 201,
    "magenta_dark": 90,
    "magenta_dark_1": 91,
    "red_1": 124,
    "red_2": 160,
    "red_3": 196,
    "indian_red": 131,
    "indian_red_1": 167,
    "indian_red_2": 203,
    "indian_red_3": 204,
    "dark_red": 52,
    "dark_red_1": 88,
    "violet_red_medium": 126,
    "violet_red_pale": 211,
    "pink": 175,
    "pink_1": 218,
    "hot_pink": 132,
    "hot_pink_1": 168,
    "hot_pink_2": 169,
    "hot_pink_3": 205,
    "hot_pink_4": 206,
    "light_pink": 95,
    "light_pink_1": 174,
    "light_pink_2": 217,
    "pink_deep": 53,
    "pink_deep_1": 89,
    "pink_deep_2": 125,
    "pink_deep_3": 161,
    "pink_deep_4": 162,
    "pink_deep_5": 197,
    "pink_deep_6": 198,
    "pink_deep_7": 199,
    "purple": 54,
    "purple_1": 55,
    "purple_2": 56,
    "purple_3": 93,
    "purple_dim": 60,
    "purple_medium": 97,
    "purple_medium_1": 98,
    "purple_medium_2": 104,
    "purple_light": 129,
    "purple_light_1": 135,
    "purple_light_2": 140,
    "purple_light_3": 141,
    "turquoise": 30,
    "turquoise_1": 45,
    "medium_turquoise": 80,
    "dark_turquoise": 44,
    "pale_turquoise": 66,
    "pale_turquoise_1": 159,
    "dark_cyan": 36,
    "cyan_1": 43,
    "cyan_2": 50,
    "cyan_3": 51,
    "cyan_4": 87,
    "cyan_5": 123,
    "cyan_light": 116,
    "cyan_light_1": 152,
    "cyan_light_2": 195,
    "chartreuse1": 64,
    "chartreuse2": 70,
    "chartreuse3": 76,
    "chartreuse4": 82,
    "chartreuse5": 112,
    "chartreuse6": 118,
    "dark_green": 22,
    "green_1": 28,
    "green_2": 34,
    "green_3": 40,
    "green_4": 46,
    "spring_green": 29,
    "spring_green_1": 35,
    "spring_green_2": 41,
    "spring_green_3": 42,
    "spring_green_4": 47,
    "spring_green_5": 48,
    "spring_green_6": 49,
    "green_light": 119,
    "green_light_2": 120,
    "pale_green": 77,
    "pale_green_1": 114,
    "pale_green_2": 121,
    "pale_green_3": 156,
    "sea_green": 78,
    "sea_green_1": 83,
    "sea_green_2": 84,
    "sea_green_3": 85,
    "sea_green_light": 37,
    "sea_green_dark": 65,
    "sea_green_dark_1": 71,
    "sea_green_dark_2": 108,
    "sea_green_dark_3": 115,
    "sea_green_dark_4": 150,
    "sea_green_dark_5": 151,
    "sea_green_dark_6": 157,
    "sea_green_dark_7": 158,
    "sea_green_dark_8": 193,
    "olive_green_dark": 107,
    "olive_green_dark_1": 113,
    "olive_green_dark_2": 149,
    "olive_green_dark_3": 155,
    "olive_green_dark_4": 191,
    "olive_green_dark_5": 192,
    "navy_blue": 17,
    "royal_blue": 63,
    "cornflower_blue": 69,
    "blue_1": 19,
    "blue_2": 20,
    "blue_3": 21,
    "dark_blue": 18,
    "dodger_blue": 26,
    "dodger_blue_1": 27,
    "dodger_blue_2": 33,
    "cadet_blue": 72,
    "cadet_blue_2": 73,
    "slate_blue": 61,
    "slate_blue_1": 62,
    "slate_blue_2": 99,
    "slate_blue_light": 105,
    "sky_blue": 74,
    "sky_blue_1": 111,
    "sky_blue_2": 117,
    "sky_blue_light": 109,
    "sky_blue_light_1": 110,
    "sky_blue_light_2": 153,
    "sky_blue_deep": 23,
    "sky_blue_deep_1": 24,
    "sky_blue_deep_2": 25,
    "sky_blue_deep_3": 31,
    "sky_blue_deep_4": 32,
    "sky_blue_deep_5": 38,
    "sky_blue_deep_6": 39,
    "steel_blue": 67,
    "steel_blue_1": 68,
    "steel_blue_2": 75,
    "steel_blue_3": 81,
    "steel_blue_light": 146,
    "steel_blue_light_1": 147,
    "steel_blue_light_2": 189,
    "slate_gray": 103,
    "grey_dim": 59,
    "grey_dark": 102,
    "grey_light": 145,
    "gainsboro": 188,
    "grey_1": 232,
    "grey_2": 233,
    "grey_3": 234,
    "grey_4": 235,
    "grey_5": 236,
    "grey_6": 237,
    "grey_7": 238,
    "grey_8": 239,
    "grey_9": 240,
    "grey_10": 241,
    "grey_11": 242,
    "grey_12": 243,
    "grey_13": 244,
    "grey_14": 245,
    "grey_15": 246,
    "grey_16": 247,
    "grey_17": 248,
    "grey_18": 249,
    "grey_19": 250,
    "grey_20": 251,
    "grey_21": 252,
    "grey_22": 253,
    "grey_23": 254,
    "grey_24": 255,
}


# Just the names of colors
STANDARD_COLORNAMES: list = list(FGCOLORS_STANDARD.keys())

COLORS256_COLORNAMES: list = list(COLORS_256.keys())


# Type Annotations
Quality = Literal["high", "medium", "low"]

RGBTuple = Tuple[int, int, int]

Attribute = Literal[
    "bold",
    "dark",
    "italic",
    "underline",
    "blink",
    "reverse",
    "concealed",
    "strike",
    "double-underline",
    "overline",
]


StandardColor = Literal[
    "black",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_grey",
    "dark_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "white",
]


Color256 = Literal[
    "black",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_grey",
    "dark_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "smokewhite",
    "white",
    "tan",
    "honeydew",
    "light_coral",
    "cornsilk",
    "plum",
    "plum_1",
    "plum_2",
    "plum_dim",
    "plum_dim_1",
    "khaki",
    "khaki_1",
    "khaki_dark",
    "yellow_1",
    "yellow_2",
    "yellow_3",
    "yellow_dark",
    "yellow_pale",
    "yellow_green",
    "yellow_green_1",
    "yellow_green_2",
    "aquamarine",
    "aquamarine_1",
    "aquamarine_2",
    "violet",
    "violet_dark",
    "violet_dark_1",
    "blue_violet",
    "orchid",
    "orchid_medium",
    "orchid_medium_1",
    "orchid_medium_2",
    "orchid_light",
    "orchid_light_1",
    "orchid_light_2",
    "gold",
    "gold_1",
    "gold_2",
    "goldenrod",
    "goldenrod_light",
    "goldenrod_light_1",
    "goldenrod_light_2",
    "goldenrod_light_3",
    "goldenrod_dark",
    "salmon",
    "salmon_dim",
    "salmon_light",
    "salmon_light_1",
    "brown_rosy",
    "brown_sandy",
    "navajo_white",
    "navajo_white_1",
    "misty_rose",
    "misty_rose_1",
    "thistle",
    "thistle_1",
    "wheat",
    "wheat_light",
    "orange",
    "orange_green",
    "orange_light_1",
    "orange_light_2",
    "orange_dim",
    "orange_dim_1",
    "orange_red_1",
    "orange_red_2",
    "magenta_1",
    "magenta_2",
    "magenta_3",
    "magenta_4",
    "magenta_light",
    "magenta_light_1",
    "magenta_dark",
    "magenta_dark_1",
    "red_1",
    "red_2",
    "red_3",
    "indian_red",
    "indian_red_1",
    "indian_red_2",
    "indian_red_3",
    "dark_red",
    "dark_red_1",
    "violet_red_medium",
    "violet_red_pale",
    "pink",
    "pink_1",
    "hot_pink",
    "hot_pink_1",
    "hot_pink_2",
    "hot_pink_3",
    "hot_pink_4",
    "light_pink",
    "light_pink_1",
    "light_pink_2",
    "pink_deep",
    "pink_deep_1",
    "pink_deep_2",
    "pink_deep_3",
    "pink_deep_4",
    "pink_deep_5",
    "pink_deep_6",
    "pink_deep_7",
    "purple",
    "purple_1",
    "purple_2",
    "purple_3",
    "purple_dim",
    "purple_medium",
    "purple_medium_1",
    "purple_medium_2",
    "purple_light",
    "purple_light_1",
    "purple_light_2",
    "purple_light_3",
    "turquoise",
    "turquoise_1",
    "medium_turquoise",
    "dark_turquoise",
    "pale_turquoise",
    "pale_turquoise_1",
    "dark_cyan",
    "cyan_1",
    "cyan_2",
    "cyan_3",
    "cyan_4",
    "cyan_5",
    "cyan_light",
    "cyan_light_1",
    "cyan_light_2",
    "chartreuse1",
    "chartreuse2",
    "chartreuse3",
    "chartreuse4",
    "chartreuse5",
    "chartreuse6",
    "dark_green",
    "green_1",
    "green_2",
    "green_3",
    "green_4",
    "spring_green",
    "spring_green_1",
    "spring_green_2",
    "spring_green_3",
    "spring_green_4",
    "spring_green_5",
    "spring_green_6",
    "green_light",
    "green_light_2",
    "pale_green",
    "pale_green_1",
    "pale_green_2",
    "pale_green_3",
    "sea_green",
    "sea_green_1",
    "sea_green_2",
    "sea_green_3",
    "sea_green_light",
    "sea_green_dark",
    "sea_green_dark_1",
    "sea_green_dark_2",
    "sea_green_dark_3",
    "sea_green_dark_4",
    "sea_green_dark_5",
    "sea_green_dark_6",
    "sea_green_dark_7",
    "sea_green_dark_8",
    "olive_green_dark",
    "olive_green_dark_1",
    "olive_green_dark_2",
    "olive_green_dark_3",
    "olive_green_dark_4",
    "olive_green_dark_5",
    "navy_blue",
    "royal_blue",
    "cornflower_blue",
    "blue_1",
    "blue_2",
    "blue_3",
    "dark_blue",
    "dodger_blue",
    "dodger_blue_1",
    "dodger_blue_2",
    "cadet_blue",
    "cadet_blue_2",
    "slate_blue",
    "slate_blue_1",
    "slate_blue_2",
    "slate_blue_light",
    "sky_blue",
    "sky_blue_1",
    "sky_blue_2",
    "sky_blue_light",
    "sky_blue_light_1",
    "sky_blue_light_2",
    "sky_blue_deep",
    "sky_blue_deep_1",
    "sky_blue_deep_2",
    "sky_blue_deep_3",
    "sky_blue_deep_4",
    "sky_blue_deep_5",
    "sky_blue_deep_6",
    "steel_blue",
    "steel_blue_1",
    "steel_blue_2",
    "steel_blue_3",
    "steel_blue_light",
    "steel_blue_light_1",
    "steel_blue_light_2",
    "slate_gray",
    "grey_dim",
    "grey_dark",
    "grey_light",
    "gainsboro",
    "grey_1",
    "grey_2",
    "grey_3",
    "grey_4",
    "grey_5",
    "grey_6",
    "grey_7",
    "grey_8",
    "grey_9",
    "grey_10",
    "grey_11",
    "grey_12",
    "grey_13",
    "grey_14",
    "grey_15",
    "grey_16",
    "grey_17",
    "grey_18",
    "grey_19",
    "grey_20",
    "grey_21",
    "grey_22",
    "grey_23",
    "grey_24",
]

# Color mode
ColorMode = Union[int, Literal[4, 8, 24]]

# Master Color Type
Color = Union[str, StandardColor, Color256, RGBTuple]



""" All Exceptions used by `Ansy` """


class StyleError(Exception):
    """Invalid style for `Ansy strings`"""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self.args = args


class ColorModeError(Exception):
    """Inappropriate color mode that is unrecognized by `Ansy`"""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self.args = args


class InvalidColorError(Exception):
    """Inappropriate color that is unrecognized by `Ansy`"""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self.args = args


class RGBError(InvalidColorError):
    """Invalid RGB"""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self.args = args


class HexError(InvalidColorError):
    """Invalid Hex color code"""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self.args = args



""" Run `pytest test_ansy.py` """

import ansy
from ansy.exceptions import *
from typing import Generator
from pytest import raises


def test_colored_basic():
    assert ansy.colored(0, "blue", force_color=True) == "\x1b[34m0\x1b[0m"

    assert ansy.colored(None, "blue", force_color=True) == ""
    assert ansy.colored("", "blue", force_color=True) == ""
    assert ansy.colored(False, "blue", force_color=True) == ""

    s = "This is string."
    # 4Bit
    assert ansy.colored(s, "blue", force_color=True) == "\x1b[34mThis is string.\x1b[0m"
    assert (
        ansy.colored(s, "blue", "dark_grey", force_color=True)
        == "\x1b[100m\x1b[34mThis is string.\x1b[0m"
    )

    # 8Bit
    assert (
        ansy.colored(s, "blue", color_mode=8, force_color=True)
        == "\x1b[38;5;4mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(s, "plum", 234, color_mode=8, force_color=True)
        == "\x1b[48;5;234m\x1b[38;5;183mThis is string.\x1b[0m"
    )

    # 24Bit
    assert (
        ansy.colored(s, (255, 100, 100), "#FFF", color_mode=24, force_color=True)
        == "\x1b[48;2;255;255;255m\x1b[38;2;255;100;100mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(s, "#FFFFFF", (50, 50, 100), color_mode=24, force_color=True)
        == "\x1b[48;2;50;50;100m\x1b[38;2;255;255;255mThis is string.\x1b[0m"
    )


def test_colored_advanced():
    assert (
        ansy.colored(
            "This is string.",
            "blue",
            attrs=["bold", "italic", "strike", "overline", "underline"],
            force_color=True,
        )
        == "\x1b[4m\x1b[53m\x1b[9m\x1b[3m\x1b[1m\x1b[34mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(
            "This is string.",
            "blue",
            "dark_grey",
            attrs=["bold", "italic", "strike", "overline", "underline"],
            force_color=True,
        )
        == "\x1b[4m\x1b[53m\x1b[9m\x1b[3m\x1b[1m\x1b[100m\x1b[34mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(
            "This is string.",
            "blue",
            color_mode=8,
            attrs=["bold", "italic", "strike", "overline", "underline"],
            force_color=True,
        )
        == "\x1b[4m\x1b[53m\x1b[9m\x1b[3m\x1b[1m\x1b[38;5;4mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(
            "This is string.",
            "plum",
            234,
            color_mode=8,
            attrs=["bold", "italic", "strike", "overline", "underline"],
            force_color=True,
        )
        == "\x1b[4m\x1b[53m\x1b[9m\x1b[3m\x1b[1m\x1b[48;5;234m\x1b[38;5;183mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(
            "This is string.",
            (255, 100, 100),
            "#FFF",
            color_mode=24,
            attrs=["bold", "italic", "strike", "overline", "underline"],
            force_color=True,
        )
        == "\x1b[4m\x1b[53m\x1b[9m\x1b[3m\x1b[1m\x1b[48;2;255;255;255m\x1b[38;2;255;100;100mThis is string.\x1b[0m"
    )
    assert (
        ansy.colored(
            "This is string.",
            "#FFFFFF",
            (50, 50, 100),
            color_mode=24,
            attrs=["bold", "italic", "strike", "overline", "underline"],
            force_color=True,
        )
        == "\x1b[4m\x1b[53m\x1b[9m\x1b[3m\x1b[1m\x1b[48;2;50;50;100m\x1b[38;2;255;255;255mThis is string.\x1b[0m"
    )


def test_colored_errors():
    s = "this is a string"
    with raises(ColorModeError):
        ansy.colored(s, "blue", force_color=True, color_mode=18)

    with raises(InvalidColorError):
        ansy.colored(s, "plum", force_color=True, color_mode=4)
    with raises(InvalidColorError):
        ansy.colored(s, bgcolor="invalid", force_color=True, color_mode=8)
    with raises(InvalidColorError):
        ansy.colored(s, "#fva", force_color=True, color_mode=24)
    with raises(InvalidColorError):
        ansy.colored(s, bgcolor=(255, 502, 10), force_color=True, color_mode=24)

    with raises(AttributeError):
        ansy.colored(s, attrs=["old"], force_color=True, color_mode=24)


def test_colored_ansy():
    style = {
        "yellow": ansy.create_style(24, fgcolor="#f0c27b"),
        "red": ansy.create_style(24, fgcolor="#ff5555"),
        "new": ansy.create_style(24, bgcolor="#ff5555", attrs=["italic", "bold"]),
        "invalid": {
            "color_mode": 54,
            "fgcolor": "343434",
            "bgcolor": None,
            "attrs": None,
        },
    }
    assert (
        ansy.colored_ansy("Welcome to @red[ansy] @yellow[:)]", style, force_color=True)
        == "Welcome to \x1b[38;2;255;85;85mansy\x1b[0m \x1b[38;2;240;194;123m:)\x1b[0m"
    )

    with raises(StyleError):
        ansy.colored_ansy("Welcome to @red[ansy] @asd[:)]", style, force_color=True)

    with raises(ColorModeError):
        ansy.colored_ansy("Welcome to @red[ansy] @invalid[:)]", style, force_color=True)


def test_colored_gradient():
    assert (
        ansy.colored_gradient("asd", (25, 25, 100), (1, 1, 1))
        == "\x1b[38;2;25;25;100ma\x1b[38;2;13;13;50ms\x1b[38;2;1;1;1md\x1b[0m"
    )
    assert (
        ansy.colored_gradient("asd", (25, 25, 100), (1, 1, 1), reverse=True)
        == "\x1b[38;2;1;1;1ma\x1b[38;2;13;13;50ms\x1b[38;2;25;25;100md\x1b[0m"
    )
    assert (
        ansy.colored_gradient("a", (25, 25, 100), (1, 1, 1))
        == "\x1b[38;2;25;25;100ma\x1b[0m"
    )
    assert (
        ansy.colored_gradient("This is quality test", "#ff22ff", "b00b1e", "high")
        == "\x1b[38;2;255;34;255mT\x1b[38;2;250;32;243mh\x1b[38;2;246;31;231mi\x1b[38;2;242;30;219ms\x1b[38;2;238;29;207m \x1b[38;2;234;27;195mi\x1b[38;2;230;26;183ms\x1b[38;2;225;25;172m \x1b[38;2;221;24;160mq\x1b[38;2;217;23;148mu\x1b[38;2;213;21;136ma\x1b[38;2;209;20;124ml\x1b[38;2;205;19;112mi\x1b[38;2;200;18;101mt\x1b[38;2;196;17;89my\x1b[38;2;192;15;77m \x1b[38;2;188;14;65mt\x1b[38;2;184;13;53me\x1b[38;2;180;12;41ms\x1b[38;2;176;11;30mt\x1b[0m"
    )
    assert (
        ansy.colored_gradient("This is quality test", "#ff22ff", "b00b1e", "medium")
        == "\x1b[38;2;250;32;243mTh\x1b[38;2;242;30;219mis\x1b[38;2;234;27;195m i\x1b[38;2;225;25;172ms \x1b[38;2;217;23;148mqu\x1b[38;2;209;20;124mal\x1b[38;2;200;18;101mit\x1b[38;2;192;15;77my \x1b[38;2;184;13;53mte\x1b[38;2;176;11;30mst\x1b[0m"
    )
    assert (
        ansy.colored_gradient("This is quality test", "#ff22ff", "b00b1e", "low")
        == "\x1b[38;2;238;29;207mThis \x1b[38;2;217;23;148mis qu\x1b[38;2;196;17;89mality\x1b[38;2;176;11;30m test\x1b[0m"
    )

    with raises(InvalidColorError):
        ansy.colored_gradient("asd", (25, 25, 1000), (1, 1, 1))

    with raises(InvalidColorError):
        ansy.colored_gradient("asd", (25, 25, 100), (1, 1, -1))

    with raises(AssertionError):
        ansy.colored_gradient("asd", (25, 25, 100), (1, 1, 1), "2")


def test_make_ansi():
    with raises(ColorModeError):
        ansy.make_ansi(bgcolor="light_red", color_mode="24")
    with raises(ColorModeError):
        ansy.make_ansi(bgcolor="light_red", color_mode=256)
    with raises(ColorModeError):
        ansy.make_ansi(bgcolor="light_red", color_mode=-4)
    with raises(AttributeError):
        ansy.make_ansi(attrs=["something", "invalid"])


def test_make_ansi_4bit():
    # Expected...
    e = "\x1b[53m\x1b[1m\x1b[100m\x1b[32m"
    # Created...
    f = ansy.make_ansi("green", "dark_grey", ["bold", "overline"], 4)
    assert f == e

    with raises(InvalidColorError):
        ansy.make_ansi(fgcolor="asdf", color_mode=4)
    with raises(InvalidColorError):
        ansy.make_ansi(fgcolor="light_ref", color_mode=4)
    with raises(InvalidColorError):
        ansy.make_ansi(fgcolor="bla", color_mode=4)
    with raises(InvalidColorError):
        ansy.make_ansi(fgcolor=33, color_mode=4)


def test_make_ansi_8bit():
    # Expected...
    e = "\x1b[53m\x1b[3m\x1b[48;5;3m\x1b[38;5;183m"
    # Created...
    f = ansy.make_ansi("plum", "yellow", ["italic", "overline"], 8)
    assert f == e

    with raises(InvalidColorError):
        ansy.make_ansi(bgcolor="asdf", color_mode=8)
    with raises(InvalidColorError):
        ansy.make_ansi(bgcolor=-5, color_mode=8)
    with raises(InvalidColorError):
        ansy.make_ansi(bgcolor=500, color_mode=8)


def test_make_ansi_24bit():
    # Expected...
    e = "\x1b[9m\x1b[8m\x1b[48;2;0;0;0m\x1b[38;2;255;100;100m"
    # Created...
    f = ansy.make_ansi((255, 100, 100), (0, 0, 0), ["concealed", "strike"], 24)
    assert f == e

    with raises(InvalidColorError):
        ansy.make_ansi(fgcolor="#asdf", color_mode=24)
    with raises(InvalidColorError):
        ansy.make_ansi(bgcolor=(255, 500, 500), color_mode=24)


def test_get_all_colors():
    # Type checking: Generator Expected
    assert isinstance(ansy.get_all_colors(), Generator)

    # Unsorted test
    for clr in ansy.get_all_colors():
        assert clr == ("black", 0)
        break

    # Sorted test
    for clr in ansy.get_all_colors(True):
        assert clr == ("aquamarine", 79)
        break


def test_de_ansi():
    # String to clean
    ansi_string = "\x1b[38;2;242;105;117mThis i\x1b[38;2;228;112;139ms an a\x1b[38;2;213;119;160mnsy te\x1b[38;2;198;126;182mst for\x1b[38;2;184;133;203m color\x1b[38;2;169;140;225ming an\x1b[38;2;154;147;246md stuf\x1b[38;2;150;150;254mf.\x1b[0m"
    clean = ansy.de_ansi(ansi_string)
    expected = "This is an ansy test for coloring and stuff."
    assert clean == expected

    with raises(TypeError):
        ansy.de_ansi(1234)

    with raises(TypeError):
        ansy.de_ansi(None)


def test_contains_ansi():
    dummy = "dummy string"
    assert not ansy.contains_ansi(ansy.colored(None))
    assert not ansy.contains_ansi(ansy.colored(""))
    assert not ansy.contains_ansi(ansy.colored(dummy))

    assert ansy.contains_ansi(
        ansy.colored(dummy, "red", "dark_grey", ["bold", "underline"], force_color=True)
    )
    assert ansy.contains_ansi(
        ansy.colored(
            dummy,
            "brown_sandy",
            "dark_grey",
            ["bold", "underline"],
            8,
            force_color=True,
        )
    )
    assert ansy.contains_ansi(
        ansy.colored(
            dummy, "#B00B1E", "#C0DDE5", ["bold", "underline"], 24, force_color=True
        )
    )

    with raises(TypeError):
        ansy.contains_ansi(None)
    with raises(TypeError):
        ansy.contains_ansi(123)


def test_make_gradient():
    grad = ansy.make_gradient((255, 100, 100), (50, 50, 100), 5, False)
    expected = [
        (255, 100, 100),
        (203, 87, 100),
        (152, 75, 100),
        (101, 62, 100),
        (50, 50, 100),
    ]

    # Colors checking
    grad_list = list(grad)
    assert len(grad_list) == len(expected)

    # Contents checking
    assert grad_list == expected

    with raises(InvalidColorError):
        list(ansy.make_gradient((1231, 123, 123), (123, 123, 123), 3))

    with raises(InvalidColorError):
        list(ansy.make_gradient((121, 123, 123), (123,), 3))

    with raises(InvalidColorError):
        list(ansy.make_gradient("fff", "#FFF123", 3))


def test_get_random_color():
    c4 = ansy.get_random_color(4)
    assert isinstance(c4, str)
    assert c4 in ansy.STANDARD_COLORNAMES

    c8 = ansy.get_random_color(8)
    assert isinstance(c8, str)
    assert c8 in ansy.COLORS256_COLORNAMES

    c24 = ansy.get_random_color(24)
    assert isinstance(c24, tuple)
    assert ansy.is_valid_rgb(c24) == True

    with raises(ColorModeError):
        ansy.get_random_color(245)


def test_create_style():
    e4 = {
        "color_mode": 4,
        "fgcolor": "light_red",
        "bgcolor": "dark_grey",
        "attrs": ["italic"],
    }
    s4 = ansy.create_style(4, "light_red", "dark_grey", ["italic"])

    assert s4 == e4

    # 8Bit checking
    e8 = {
        "color_mode": 8,
        "fgcolor": "smokewhite",
        "bgcolor": "plum",
        "attrs": ["bold", "underline"],
    }
    s8 = ansy.create_style(8, "smokewhite", "plum", ["bold", "underline"])

    assert s8 == e8

    # 24Bit checking
    e24 = {
        "color_mode": 24,
        "fgcolor": (255, 100, 100),
        "bgcolor": (0, 1, 0),
        "attrs": None,
    }
    s24 = ansy.create_style(24, (255, 100, 100), (0, 1, 0))

    assert s24 == e24

    with raises(InvalidColorError):
        ansy.create_style(8, "invcalid")

    with raises(InvalidColorError):
        ansy.create_style(4, (55, 100, 100), "123123")

    with raises(ColorModeError):
        ansy.create_style(2, (255, 100, 100), (0, 1, 0))

    with raises(AttributeError):
        ansy.create_style(24, (255, 100, 100), (0, 1, 0), ["boldd"])


def test_search_colors():
    expected = [
        ("plum", 183),
        ("plum_1", 176),
        ("plum_2", 219),
        ("plum_dim", 96),
        ("plum_dim_1", 139),
    ]

    result = ansy.search_colors("plum")

    assert isinstance(result, Generator)

    assert list(result) == expected


def test_hex_to_rgb():
    expected = (255, 242, 34)
    assert ansy.hex_to_rgb("#FFF222") == expected

    invalid_values = ["#FFF2", "$abc123", 123, None, 1.1, True, False]
    for v in invalid_values:
        with raises(HexError):
            ansy.hex_to_rgb(v)


def test_rgb_to_hex():
    expected = "#FFF222"
    assert ansy.rgb_to_hex((255, 242, 34)) == expected

    invalid_values = [
        ("123", "123", "12"),
        "('123','123', '12')",
        123,
        None,
        1.1,
        True,
        False,
    ]
    for v in invalid_values:
        with raises(RGBError):
            ansy.rgb_to_hex(v)


def test_colorname_to_code():
    assert ansy.colorname_to_code("plum") == ansy.COLORS_256["plum"]
    assert ansy.colorname_to_code(143) == 143
    assert ansy.colorname_to_code(None) == None

    # Raises no errors whatsoever!
    for v in ("asdf", 123, None, True, False, min, (12, 12), [123, 123], {"asd": 123}):
        ansy.colorname_to_code(v)


def test_code_to_colorname():
    assert ansy.code_to_colorname(ansy.COLORS_256["plum"]) == "plum"
    assert ansy.code_to_colorname("plum") == "plum"
    assert ansy.code_to_colorname(None) == None

    for v in ("asdf", 123, None, True, False, min):
        ansy.code_to_colorname(v)


# Tests for all validators


def test_is_valid_hex():
    # Valid Hex Checker
    valids = ["FFF", "#FFF", "fffFFF", "#FFFFFF"]
    for v in valids:
        assert ansy.is_valid_hex(v) == True

    invalids = ["#as", "#FFFFFz", None, 123]
    for v in invalids:
        assert ansy.is_valid_hex(v) == False


def test_is_valid_rgb():
    valids = [(200, 100, 1), (0, 0, 0), (255, 255, 255)]
    for v in valids:
        assert ansy.is_valid_rgb(v) == True

    invalids = [(), (255, 1), (-50, -5, 255), (500, 1000, 100), (None), (123), ("123")]
    for v in invalids:
        assert ansy.is_valid_rgb(v) == False


def test_is_valid_attr():
    # Valid Attr checker
    for attr in ("bold", "italic", "strike", "underline"):
        assert ansy.is_valid_attr(attr) == True

    # Invalid values
    for v in ("asd", 123, None, True, False):
        assert ansy.is_valid_attr(v) == False


def test_is_valid_color():
    valids = [
        ("yellow", 4),
        ("aquamarine", 8),
        ("yellow", 8),
        ("#fff", 24),
        ((255, 255, 100), 24),
    ]
    for v in valids:
        assert ansy.is_valid_color(*v) == True

    invalids = [
        ("aquamarine", 4),
        ("aquamarine", 24),
        ("yellow", 24),
        ("#fff", 4),
        ("#fff", 8),
        ("#safsaf", 4),
        ("#safsaf", 8),
        ("#safsaf", 24),
        ((255,), 4),
        ((255,), 8),
        ((255,), 24),
        ((255, 255, 100), 4),
        ((255, 255, 100), 8),
    ]
    for v in invalids:
        assert ansy.is_valid_color(*v) == False


if __name__ == "__main__":
    test_contains_ansi()
